/////////////////////////////////////////////////////////////////////////////
///
/// @file datatext.h
///
/// Converting a binary data buffer into a printable text and vice versa.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on November 12, 2013. Last revision: November 12, 2013.


#ifndef _CMGRAPHS_DATATEXT_H_
#define _CMGRAPHS_DATATEXT_H_


// include some standard C++ header files
#include <string>

// include all the necessary header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/multiwork/mwdata.h"


// --------------------------------------------------
// ---------- binary data - printable text ----------
// --------------------------------------------------

/// Encodes the entire binary data into a text string (single line,
/// printable characters only).
inline std::string data2text (const chomp::multiwork::mwData &data)
{
	const char *buffer (data. Buffer ());
	int length (data. Length ());
	std::string str;
	str. resize (8 + 2 * length);

	// encode the length of the buffer
	for (int i = 0; i < 4; ++ i)
	{
		int byte ((length & (0xFF << (i << 3))) >> (i << 3));
		str [i << 1] = 'A' + (byte & 0x0F);
		str [(i << 1) + 1] = 'G' + ((byte >> 4) & 0x0F);
	}

	// encode the buffer
	for (int i = 0; i < length; ++ i)
	{
		str [8 + (i << 1)] = 'a' + (buffer [i] & 0x0F);
		str [9 + (i << 1)] = 'g' + (((buffer [i]) >> 4) & 0x0F);
	}

	return str;
} /* data2text */

/// Decodes a prevously prepared text back into binary data.
/// Resets the data and replaces its contents with the decoded buffer.
/// Throws an exception in case of errors.
inline void text2data (const std::string &str,
	chomp::multiwork::mwData &data)
{
	if (str. size () < 8)
		throw "Trying to decode a string that is too short.";

	// decode the length of the data buffer
	int length (0);
	for (int i = 0; i < 4; ++ i)
	{
		int b1 = str [i << 1] - 'A';
		int b2 = str [(i << 1) + 1] - 'G';
		if ((b1 < 0) || (b1 > 0x0F) || (b2 < 0) || (b2 > 0x0F))
			throw "Wrong length while decoding a string.";
		length |= (b1 | (b2 << 4)) << (i << 3);
	}
	if (length < 0)
		throw "Negative length while decoding a string.";

	// reset the data to prepare it for a new binary buffer
	data. Reset ();
	if (!length)
		return;

	// allocate the data buffer if necessary
	unsigned char *buffer = new unsigned char [length];
	data. Take (buffer, length);

	// decode the data buffer
	for (int i = 0; i < length; ++ i)
	{
		int b1 = str [8 + (i << 1)] - 'a';
		int b2 = str [9 + (i << 1)] - 'g';
		if ((b1 < 0) || (b1 > 0x0F) || (b2 < 0) || (b2 > 0x0F))
			throw "Wrong contents while decoding a string.";
		buffer [i] = b1 | (b2 << 4);
	}

	return;
} /* text2data */


#endif // _CMGRAPHS_DATATEXT_H_

