/////////////////////////////////////////////////////////////////////////////
///
/// @file m_henon.h
///
/// The Henon map.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on January 18, 2010. Last revision: January 18, 2010.


#ifndef _CMGRAPHS_HENON_H_
#define _CMGRAPHS_HENON_H_


// the generic class for the type of a map
#include "maptype.h"

// the interval arithmetic data types
#include "typeintv.h"


// --------------------------------------------------
// -------- a sample difference equation map --------
// --------------------------------------------------

/// This class defines the Henon map.
/// There are only two parameters of this map: 'a' and 'b'.
class MapHenon: public MapType
{
public:
	/// The default constructor.
	MapHenon ();

	/// Computes the image of a box whose left and right coordinates
	/// are given. Fills in the images with the left and right
	/// coordinates of the image box.
	void compute (const double *xleft, const double *xright,
		double *yleft, double *yright, int dim,
		const spcCoord *coord, int subdiv) const;

}; /* class MapHenon */

// --------------------------------------------------

inline MapHenon::MapHenon ()
{
	testIntervals ();
	return;
} /* MapHenon::MapHenon */

inline void MapHenon::compute (const double *xleft, const double *xright,
	double *yleft, double *yright, int dim, const spcCoord *, int) const
{
	if (dim != 2)
		throw "Wrong phase space dim for the Henon map.";
	IntervalType a (getLeftParam (0), getRightParam (0));
	IntervalType b (getLeftParam (1), getRightParam (1));
	IntervalType x (xleft [0], xright [0]);
	IntervalType x2;
	if ((xleft [0] >= 0) || (xright [0] <= 0))
	{
		x2 = x * x;
	}
	else
	{
		IntervalType half (0, (-xleft [0] >= xright [0]) ?
			(-xleft [0]) : xright [0]);
		x2 = half * half;
	}
	IntervalType y (xleft [1], xright [1]);
	IntervalType one (1.0, 1.0);
	IntervalType xNew (y + one - a * x2);
	IntervalType yNew (b * x);
	yleft [0] = xNew. leftBound ();
	yright [0] = xNew. rightBound ();
	yleft [1] = yNew. leftBound ();
	yright [1] = yNew. rightBound ();

	// swich the rounding direction to the neutral one
	resetRounding ();
	return;
} /* MapHenon::compute */


#endif // _CMGRAPHS_HENON_H_

