/////////////////////////////////////////////////////////////////////////////
///
/// @file p_neuron09.h
///
/// Parameters for for the Chialvo neuron model in R^2
/// with 2 varying parameters.
/// This is a configuration file for the program "cmgraphs.cpp"
/// which defines the ranges of parameters to scan
/// and the details of the phase space.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2022 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on February 9, 2008. Last revision: July 18, 2022.


#ifndef _CMGRAPHS_P_H_
#define _CMGRAPHS_P_H_


namespace custom {

/// Parameters for sample computations for the 2-dimensional Leslie
/// population model with 2 varying parameters.
namespace neuron09 {


// --------------------------------------------------
// --------- the phase space configuration ----------
// --------------------------------------------------

/// The dimension of the phase space.
const int spaceDim = 2;

/// Is space wrapping (a.k.a. periodic boundary conditions)
/// in effect in the given directions?
const bool spaceWrapping [spaceDim] = {false, false};

/// The initial depth of subdivisions in the phase space.
const int initialDepth = 7;

/// The final depth of subdivisions in the phase space.
const int finalDepth = 10;


// --------------------------------------------------
// ------------ the parameter selection -------------
// --------------------------------------------------

/// The dimension of the parameter space to iterate. Only those parameters
/// which are varying are taken into consideration here.
const int paramDim = 2;

/// The numbers of subintervals in each direction of the parameter space.
/// The type of these numbers must match "parCoord" in "typedefs.h".
const short int paramSubdiv [paramDim] = {100, 100};

/// The numbers of parameters to subdivide.
const int paramSelect [paramDim] = {1, 3};


// --------------------------------------------------
// ---------- the parameter space settings ----------
// --------------------------------------------------

/// The number of all the parameters, both varying and fixed.
/// The four parameters are: a, b, c, k.
const int paramCount = 4;

/// The left bounds on the parameters.
const double paramLeft [paramCount] = {0.89, 0.0, 0.28, 0.0};

/// The right bounds on the parameters.
const double paramRight [paramCount] = {0.89, 1.0, 0.28, 0.2};


// --------------------------------------------------
// -------------- the phase space size --------------
// --------------------------------------------------

/// An array with the offset of the rectangular area
/// in the phase space which contains the invariant sets of interest.
const double spaceOffset [2] = {-0.1, -5};

/// An array with the width of the rectangular area
/// in the phase space which contains the invariant sets of interest.
const double spaceWidth [2] = {9.1, 8};


// --------------------------------------------------
// ------------------ refinements -------------------
// --------------------------------------------------

/// The number of refinements that should be done if a Morse set
/// with the trivial index is encountered or if the image of a Morse set
/// sticks out of the rectangular region to which the computations
/// are restricted. Set to zero for no refinements.
const int refineDepth = 5;

/// The maximal allowed size of a set of cubes in the phase space
/// which can be refined at the initial subdivision level.
/// Set to zero to suppress any refinements.
const int maxRefineSize0 = 100000;

/// The maximal allowed size of a set of cubes in the phase space
/// which can be refined at the subsequent subdivision levels.
/// Set to zero to suppress any refinements.
const int maxRefineSize1 = 40000;


// --------------------------------------------------
// ------------------- map images -------------------
// --------------------------------------------------

/// The maximal allowed diameter of the cubical image of a signle box.
/// This limit is set up in order to prevent a blow-up of the program
/// in the memory because of accumulating huge images of boxes.
const int maxImageDiameter = 2000;

/// The maximal allowed volume of the cubical image of a single box.
/// This limit is set up in order to prevent a blow-up of the program
/// in the memory because of accumulating huge images of boxes.
const int maxImageVolume = 4000;


// --------------------------------------------------
// -------------- max index pair size ---------------
// --------------------------------------------------

/// The maximal allowed size of the index pair. This size is measured
/// as the total number of cubes in the index pair and its forward image.
const int maxIndexPairSize = 1200000;


// --------------------------------------------------
// --------------- joining Morse sets ---------------
// --------------------------------------------------

/// The maximal number of cubes in a trivial Morse set for which an attempt
/// is made to join this set with another near-by Morse set.
/// Set to zero to suppress joining.
const int maxJoinSize = 10;

/// The maximal size of a connecting orbit between two Morse sets
/// which can be considered for joining.
const int maxJoinConnection = 1000;

/// The maximal allowed distance between two Morse sets which can be
/// considered for joining. Note: It doesn't work with space wrapping.
const int maxJoinDistance = 10;


// --------------------------------------------------
// ----------------- other settings -----------------
// --------------------------------------------------

/// Should the ordering between the Morse sets be taken into consideration
/// while determining whether two Morse decompositions
/// computed for adjacent parameter boxes are in the same class?
const bool compareMorseOrdering = true;

/// Ignoring the isolation problem while matching Morse decompositions.
/// If this constant is set to "true" then Morse sets are matched
/// with each other indepent of whether the Conley index could be computed
/// for them or not. Otherwise, 'correct' and 'wrong' Morse sets
/// are matched with their counterparts in the same category only.
const bool ignoreIsolationForContinuation = true;

/// Ignoring the isolation problem while computing the Conley index.
/// If the constant is set to "true" then the index pair constructed on the
/// basis of a combinatorial Morse set with respect to the dynamics
/// restricted to the rectangular area of interest is also restricted
/// to that area by means of projecting all the boxes that stick out
/// of the region onto a single layer of boxes around the boundary.
const bool ignoreIsolationForConleyIndex = false;


} // namespace neuron09
} // namespace custom


#endif // _CMGRAPHS_P_H_

