/////////////////////////////////////////////////////////////////////////////
///
/// @file typeintv.h
///
/// Data types for interval arithmetic.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on October 20, 2008. Last revision: June 29, 2012.


#ifndef _CMGRAPHS_TYPEINTV_H_
#define _CMGRAPHS_TYPEINTV_H_


#ifdef __USE_OLD_CAPD__

// the interval arithmetic package from the CAPD library
#include "capd/interval/DoubleInterval.h"

/// The type of an interval (from the CAPD library 2.0).
typedef capd::intervals::DoubleInterval IntervalType;

#else

// the interval arithmetic package from the CAPD library
#include "capd/intervals/lib.h"
#include "capd/rounding/DoubleRounding.h"

/// The type of an interval (from the CAPD library 2.9/3.0 beta).
typedef capd::DInterval IntervalType;

#endif

// include selected header files from the CHomP library
#include "chomp/system/textfile.h"


// --------------------------------------------------
// ------ resetting the rounding if necessary -------
// --------------------------------------------------

/// This function resets rounding switches of the processor
/// and sets rounding to the nearest.
/// It is necessary in some cases, e.g., if the CAPD leaves
/// incorrect rounding mode after doing some interval operations.
inline void resetRounding ()
{
	capd::rounding::DoubleRounding::roundNearest ();
	return;
} /* resetRounding */


// --------------------------------------------------
// -------------- additional functions --------------
// --------------------------------------------------

/// Testing interval arithmetic.
/// This simple function runs a very basic test on whether
/// the interval arithmetic and the rounding directions work correctly.
/// It returns 'true' if everything seems to be fine,
/// or throws an exception (if requested to) or just returns 'false'
/// if there is something obviously wrong with the interval arithmetic.
inline bool testIntervals (bool throwException = false)
{
const char *warningMessage = "\
* * * * * * * * * * * * * * * * WARNING * * * * * * * * * * * * * * * *\n\
* It appears that switching the rounding mode in interval arithmetic\n\
* doesn't work correctly. As a consequence, the results of computations\n\
* will be inaccurate. In order to conduct mathematically correct rigorous\n\
* computations, you should compile both the CAPD library and this software\n\
* without optimization (remove the '-O2' switch from the makefile).\n\
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\n";
	IntervalType x (1.0, 1.0);
	IntervalType x3 = x / 3;
	resetRounding ();
	if (x3. leftBound () < x3. rightBound ())
		return true;
	static bool warned = false;
	if (throwException)
		throw "Incorrect rounding in interval arithmetic.";
	else if (!warned)
	{
		using chomp::homology::sout;
		sout << warningMessage;
		warned = true;
	}
	return false;
} /* testIntervals */


#endif // _CMGRAPHS_TYPEINTV_H_

